@@ -1,4 +1,25 @@
 /*
+ * Copyright (C) 2002 by the Massachusetts Institute of Technology.
+ * All rights reserved.
+ *
+ * Export of this software from the United States of America may
+ *   require a specific license from the United States Government.
+ *   It is the responsibility of any person or organization contemplating
+ *   export to obtain such a license before exporting.
+ * 
+ * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
+ * distribute this software and its documentation for any purpose and
+ * without fee is hereby granted, provided that the above copyright
+ * notice appear in all copies and that both that copyright notice and
+ * this permission notice appear in supporting documentation, and that
+ * the name of M.I.T. not be used in advertising or publicity pertaining
+ * to distribution of the software without specific, written prior
+ * permission.  Furthermore if you modify this software you must label
+ * your software as modified software and not distribute it in such a
+ * fashion that it might be confused with the original M.I.T. software.
+ * M.I.T. makes no representations about the suitability of
+ * this software for any purpose.  It is provided "as is" without express
+ * or implied warranty.
  * Copyright 1994 by OpenVision Technologies, Inc.
  * 
  * Permission to use, copy, modify, distribute, and sell this software
@@ -22,8 +43,11 @@
 
 #include <krb5.h>
 #include <kadm5/admin.h>
+#include <krb5/adm_proto.h>
 #include <com_err.h>
+#include <stdarg.h>
 
+#include <assert.h>
 #include <stdio.h>
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
@@ -36,7 +60,11 @@
 #include <netinet/in.h>
 
 #include <krb.h>
-#include "krb524.h"
+#include "krb524d.h"
+
+#if defined(NEED_DAEMON_PROTO)
+extern int daemon(int, int);
+#endif
 
 #define TIMEOUT 60
 #define TKT_BUFSIZ 2048
@@ -45,39 +73,45 @@
 char *whoami;
 int signalled = 0;
 static int debug = 0;
-void *handle;
+void *handle = NULL;
 
 int use_keytab, use_master;
+int allow_v4_crossrealm = 0;
 char *keytab = NULL;
 krb5_keytab kt;
 
-void init_keytab(), init_master(), cleanup_and_exit();
-krb5_error_code do_connection(), lookup_service_key(), kdc_get_server_key();
+void init_keytab(krb5_context), 
+    init_master(krb5_context, kadm5_config_params *),
+    cleanup_and_exit(int, krb5_context);
+krb5_error_code do_connection(int, krb5_context);
+krb5_error_code lookup_service_key(krb5_context, krb5_principal, 
+				   krb5_enctype, krb5_kvno, 
+				   krb5_keyblock *, krb5_kvno *);
+krb5_error_code  kdc_get_server_key(krb5_context, krb5_principal, 
+				    krb5_keyblock *, krb5_kvno *,
+				    krb5_enctype, krb5_kvno);
 
-void usage(context)
+static krb5_error_code
+handle_classic_v4 (krb5_context context, krb5_ticket *v5tkt,
+		   struct sockaddr_in *saddr,
+		   krb5_data *tktdata, krb5_kvno *v4kvno);
+static krb5_error_code 
+afs_return_v4(krb5_context, const krb5_principal , int *use_v5);
+
+static void usage(context)
      krb5_context context;
 {
-     fprintf(stderr, "Usage: %s [-k[eytab]] [-m[aster] [-r realm]] [-nofork]\n", whoami);
+     fprintf(stderr, "Usage: %s [-k[eytab]] [-m[aster] [-r realm]] [-nofork] [-p portnum]\n", whoami);
      cleanup_and_exit(1, context);
 }
 
-RETSIGTYPE request_exit(signo)
+static RETSIGTYPE request_exit(signo)
      int signo;
 {
      signalled = 1;
 }
 
-#if 0
-/* this is in the kadm5 library */
-int krb5_free_keyblock_contents(context, key)
-     krb5_context context;
-     krb5_keyblock *key;
-{
-     memset(key->contents, 0, key->length);
-     krb5_xfree(key->contents);
-     return 0;
-}
-#endif
+int (*encode_v4tkt)(KTEXT, char *, unsigned int *) = 0;
 
 int main(argc, argv)
      int argc;
@@ -91,21 +125,41 @@ int main(argc, argv)
      krb5_context context;
      krb5_error_code retval;
      kadm5_config_params config_params;
+     unsigned long port = 0;
+
+     whoami = ((whoami = strrchr(argv[0], '/')) ? whoami + 1 : argv[0]);
 
      retval = krb5_init_context(&context);
      if (retval) {
-	     com_err(argv[0], retval, "while initializing krb5");
+	     com_err(whoami, retval, "while initializing krb5");
 	     exit(1);
      }
 
-     whoami = ((whoami = strrchr(argv[0], '/')) ? whoami + 1 : argv[0]);
+     {
+	 krb5int_access k5int;
+	 retval = krb5int_accessor(&k5int, KRB5INT_ACCESS_VERSION);
+	 if (retval != 0) {
+	     com_err(whoami, retval,
+		     "while accessing krb5 library internal support");
+	     exit(1);
+	 }
+	 encode_v4tkt = k5int.krb524_encode_v4tkt;
+	 if (encode_v4tkt == NULL) {
+	     com_err(whoami, 0,
+		     "krb4 support disabled in krb5 support library");
+	     exit(1);
+	 }
+     }
 
      argv++; argc--;
      use_master = use_keytab = nofork = 0;
      config_params.mask = 0;
      
      while (argc) {
-	  if (strncmp(*argv, "-k", 2) == 0)
+       if (strncmp(*argv, "-X", 2) == 0) {
+	 allow_v4_crossrealm = 1;
+       }
+       else if (strncmp(*argv, "-k", 2) == 0)
 	       use_keytab = 1;
 	  else if (strncmp(*argv, "-m", 2) == 0)
 	       use_master = 1;
@@ -118,6 +172,24 @@ int main(argc, argv)
 	       config_params.mask |= KADM5_CONFIG_REALM;
 	       config_params.realm = *argv;
 	  }
+	  else if (strcmp(*argv, "-p") == 0) {
+	      char *endptr = 0;
+	      argv++; argc--;
+	      if (argc == 0)
+		  usage (context);
+	      if (port != 0) {
+		  com_err (whoami, 0,
+			   "port number may only be specified once");
+		  exit (1);
+	      }
+	      port = strtoul (*argv, &endptr, 0);
+	      if (*endptr != '\0' || port > 65535 || port == 0) {
+		  com_err (whoami, 0,
+			   "invalid port number %s, must be 1..65535\n",
+			   *argv);
+		  exit (1);
+	      }
+	  }
 	  else
 	       break;
 	  argv++; argc--;
@@ -132,6 +204,8 @@ int main(argc, argv)
      signal(SIGHUP, SIG_IGN);
      signal(SIGTERM, request_exit);
 
+     krb5_klog_init(context, "krb524d", whoami, !nofork);
+
      if (use_keytab)
 	  init_keytab(context);
      if (use_master)
@@ -140,12 +214,16 @@ int main(argc, argv)
      memset((char *) &saddr, 0, sizeof(struct sockaddr_in));
      saddr.sin_family = AF_INET;
      saddr.sin_addr.s_addr = INADDR_ANY;
-     serv = getservbyname(KRB524_SERVICE, "udp");
-     if (serv == NULL) {
-	  com_err(whoami, 0, "service entry not found, using %d", KRB524_PORT);
-	  saddr.sin_port = htons(KRB524_PORT);
+     if (port == 0) {
+	 serv = getservbyname(KRB524_SERVICE, "udp");
+	 if (serv == NULL) {
+	     com_err(whoami, 0, "service entry `%s' not found, using %d",
+		     KRB524_SERVICE, KRB524_PORT);
+	     saddr.sin_port = htons(KRB524_PORT);
+	 } else
+	     saddr.sin_port = serv->s_port;
      } else
-	  saddr.sin_port = serv->s_port;
+	 saddr.sin_port = htons(port);
 	  
      if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
 	  com_err(whoami, errno, "creating main socket");
@@ -157,10 +235,10 @@ int main(argc, argv)
 	  cleanup_and_exit(1, context);
      }
      if (!nofork && daemon(0, 0)) {
-	  com_err(whoami, errno, "while detaching from tty");
-	  cleanup_and_exit(1, context);
+	 com_err(whoami, errno, "while detaching from tty");
+	 cleanup_and_exit(1, context);
      }
-     
+
      while (1) {
 	  FD_ZERO(&rfds);
 	  FD_SET(s, &rfds);
@@ -198,10 +276,11 @@ void cleanup_and_exit(ret, context)
      int ret;
      krb5_context context;
 {
-     if (use_master) {
+     if (use_master && handle) {
 	  (void) kadm5_destroy(handle);
      }
      if (use_keytab && kt) krb5_kt_close(context, kt);
+     krb5_klog_close(context);
      krb5_free_context(context);
      exit(ret);
 }
@@ -248,21 +327,18 @@ krb5_error_code do_connection(s, context)
 {
      struct sockaddr saddr;
      krb5_ticket *v5tkt = 0;
-     KTEXT_ST v4tkt;
-     krb5_keyblock v5_service_key, v4_service_key;
      krb5_data msgdata, tktdata;
      char msgbuf[MSGSIZE], tktbuf[TKT_BUFSIZ], *p;
-     int n, ret, saddrlen;
+     int ret, saddrlen;
+     krb5_int32 n; /* Must be 4 bytes */
      krb5_kvno v4kvno;
 
-     /* Clear out keyblock contents so we don't accidentally free the stack.*/
-     v5_service_key.contents = v4_service_key.contents = 0;
-
      msgdata.data = msgbuf;
      msgdata.length = MSGSIZE;
-
+     tktdata.data = tktbuf;
+     tktdata.length = TKT_BUFSIZ;
      saddrlen = sizeof(struct sockaddr);
-     ret = recvfrom(s, msgdata.data, msgdata.length, 0, &saddr, &saddrlen);
+     ret = recvfrom(s, msgdata.data, (int) msgdata.length, 0, &saddr, &saddrlen);
      if (ret < 0) {
        /* if recvfrom fails, we probably don't have a valid saddr to 
 	  use for the reply, so don't even try to respond. */
@@ -283,7 +359,7 @@ krb5_error_code do_connection(s, context)
 	    break;
 	  default:
 	    /* try and recognize our own error packet */
-	    if (msgdata.length == sizeof(int))
+	    if (msgdata.length == sizeof(krb5_int32))
 	      return KRB5_BADMSGTYPE;
 	    else
 	      goto error;
@@ -292,67 +368,57 @@ krb5_error_code do_connection(s, context)
      if (debug)
 	  printf("V5 ticket decoded\n");
      
-     if ((ret = lookup_service_key(context, v5tkt->server,
-				   v5tkt->enc_part.enctype,
-				   v5tkt->enc_part.kvno,
-				   &v5_service_key, NULL)))
-	  goto error;
-
-     if ((ret = lookup_service_key(context, v5tkt->server,
-				   ENCTYPE_DES3_CBC_RAW,
-				   0, /* highest kvno */
-				   &v4_service_key, &v4kvno)) &&
-	 (ret = lookup_service_key(context, v5tkt->server,
-				   ENCTYPE_LOCAL_DES3_HMAC_SHA1,
-				   0,
-				   &v4_service_key, &v4kvno)) &&
-	 (ret = lookup_service_key(context, v5tkt->server,
-				   ENCTYPE_DES3_CBC_SHA1,
-				   0,
-				   &v4_service_key, &v4kvno)) &&
-	 (ret = lookup_service_key(context, v5tkt->server,
-				   ENCTYPE_DES_CBC_CRC,
-				   0,
-				   &v4_service_key, &v4kvno)))
-	 goto error;
-
-     if (debug)
-	  printf("service key retrieved\n");
-
-     ret = krb524_convert_tkt_skey(context, v5tkt, &v4tkt, &v5_service_key,
-				   &v4_service_key,
-				   (struct sockaddr_in *)&saddr);
-     if (ret)
-	  goto error;
-
-     if (debug)
-	  printf("credentials converted\n");
-
-     tktdata.data = tktbuf;
-     tktdata.length = TKT_BUFSIZ;
-     ret = encode_v4tkt(&v4tkt, tktdata.data, &tktdata.length);
-     if (ret)
-	  goto error;
-     if (debug)
-	  printf("v4 credentials encoded\n");
-
-error:
+     if( krb5_princ_size(context, v5tkt->server) >= 1
+	 &&krb5_princ_component(context, v5tkt->server, 0)->length == 3
+	 &&strncmp(krb5_princ_component(context, v5tkt->server, 0)->data,
+		   "afs", 3) == 0) {
+	 krb5_data *enc_part;
+	 int use_v5;
+	 if ((ret = afs_return_v4(context, v5tkt->server,
+				  &use_v5)) != 0) 
+	     goto error;
+	 if ((ret = encode_krb5_enc_data( &v5tkt->enc_part, &enc_part)) != 0) 
+	     goto error;
+	 if (!(use_v5 )|| enc_part->length >= 344) {
+	     krb5_free_data(context, enc_part);
+	     if ((ret = handle_classic_v4(context, v5tkt,
+					 (struct sockaddr_in *) &saddr, &tktdata,
+					 &v4kvno)) != 0)
+		 goto error;
+	 } else {
+	   KTEXT_ST fake_v4tkt;
+	   fake_v4tkt.mbz = 0;
+	   fake_v4tkt.length = enc_part->length;
+	   memcpy(fake_v4tkt.dat, enc_part->data, enc_part->length);
+	     v4kvno = (0x100-0x2b); /*protocol constant indicating  v5
+				     * enc part only*/
+	     krb5_free_data(context, enc_part);
+	     ret = encode_v4tkt(&fake_v4tkt, tktdata.data, &tktdata.length);
+	 }
+     } else {
+	 if ((ret = handle_classic_v4(context, v5tkt,
+				     (struct sockaddr_in *) &saddr, &tktdata,
+				     &v4kvno)) != 0)
+	     goto error;
+     }
+     
+	error:
      /* create the reply */
      p = msgdata.data;
      msgdata.length = 0;
      
      n = htonl(ret);
-     memcpy(p, (char *) &n, sizeof(int));
-     p += sizeof(int);
-     msgdata.length += sizeof(int);
+     memcpy(p, (char *) &n, sizeof(krb5_int32));
+     p += sizeof(krb5_int32);
+     msgdata.length += sizeof(krb5_int32);
 
      if (ret)
 	  goto write_msg;
 
      n = htonl(v4kvno);
-     memcpy(p, (char *) &n, sizeof(int));
-     p += sizeof(int);
-     msgdata.length += sizeof(int);
+     memcpy(p, (char *) &n, sizeof(krb5_int32));
+     p += sizeof(krb5_int32);
+     msgdata.length += sizeof(krb5_int32);
 
      memcpy(p, tktdata.data, tktdata.length);
      p += tktdata.length;
@@ -360,17 +426,12 @@ error:
 
 write_msg:
      if (ret)
-	  (void) sendto(s, msgdata.data, msgdata.length, 0, &saddr, saddrlen);
+	  (void) sendto(s, msgdata.data, (int) msgdata.length, 0, &saddr, saddrlen);
      else
 	  if (sendto(s, msgdata.data, msgdata.length, 0, &saddr, saddrlen)<0)
 	       ret = errno;
      if (debug)
 	  printf("reply written\n");
-/* If we have keys to clean up, do so.*/
-     if (v5_service_key.contents)
-       krb5_free_keyblock_contents(context, &v5_service_key);
-     if (v4_service_key.contents)
-       krb5_free_keyblock_contents(context, &v4_service_key);
      if (v5tkt)
        krb5_free_ticket(context, v5tkt);
      
@@ -423,9 +484,15 @@ krb5_error_code kdc_get_server_key(context, service, key, kvnop, ktype, kvno)
     kadm5_principal_ent_rec server;
     
     if ((ret = kadm5_get_principal(handle, service, &server,
-				   KADM5_KEY_DATA)))
+				   KADM5_KEY_DATA|KADM5_ATTRIBUTES)))
 	 return ret;
 
+    if (server.attributes & KRB5_KDB_DISALLOW_ALL_TIX
+	|| server.attributes & KRB5_KDB_DISALLOW_SVR) {
+	kadm5_free_principal_ent(handle, &server);
+	return KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN;
+    }
+
     /*
      * We try kadm5_decrypt_key twice because in the case of a
      * ENCTYPE_DES_CBC_CRC key, we prefer to find a krb4 salt type
@@ -453,3 +520,112 @@ krb5_error_code kdc_get_server_key(context, service, key, kvnop, ktype, kvno)
     kadm5_free_principal_ent(handle, &server);
     return ret;
 }
+
+/*
+ * We support two  kinds of v4 credentials.  There are real v4
+ *   credentials, and  a Kerberos v5 enc part masquerading as a krb4
+ *  credential to be used by modern AFS implementations; this function
+ *  handles the classic v4 case.
+ */
+
+static krb5_error_code
+handle_classic_v4 (krb5_context context, krb5_ticket *v5tkt,
+		   struct sockaddr_in *saddr,
+		   krb5_data *tktdata, krb5_kvno *v4kvno)
+{
+    krb5_error_code ret;
+    krb5_keyblock v5_service_key, v4_service_key;
+     KTEXT_ST v4tkt;
+
+    v5_service_key.contents = NULL;
+    v4_service_key.contents = NULL;
+    
+             if ((ret = lookup_service_key(context, v5tkt->server,
+				   v5tkt->enc_part.enctype,
+				   v5tkt->enc_part.kvno,
+				   &v5_service_key, NULL)))
+	  goto error;
+
+     if ( (ret = lookup_service_key(context, v5tkt->server,
+				   ENCTYPE_DES_CBC_CRC,
+				   0,
+				   &v4_service_key, v4kvno)))
+	 goto error;
+
+     if (debug)
+	  printf("service key retrieved\n");
+     if ((ret = krb5_decrypt_tkt_part(context, &v5_service_key, v5tkt))) {
+       goto error;
+     }
+
+    if (!(allow_v4_crossrealm || krb5_realm_compare(context, v5tkt->server,
+						    v5tkt->enc_part2->client))) {
+ret =  KRB5KDC_ERR_POLICY ;
+ goto error;
+    }
+    krb5_free_enc_tkt_part(context, v5tkt->enc_part2);
+    v5tkt->enc_part2= NULL;
+
+         ret = krb524_convert_tkt_skey(context, v5tkt, &v4tkt, &v5_service_key,
+				   &v4_service_key,
+				   (struct sockaddr_in *)saddr);
+     if (ret)
+	  goto error;
+
+     if (debug)
+	  printf("credentials converted\n");
+
+     ret = encode_v4tkt(&v4tkt, tktdata->data, &tktdata->length);
+     if (ret)
+	  goto error;
+     if (debug)
+	  printf("v4 credentials encoded\n");
+
+ error:
+     if (v5tkt->enc_part2)
+	 krb5_free_enc_tkt_part(context, v5tkt->enc_part2);
+
+     if(v5_service_key.contents)
+       krb5_free_keyblock_contents(context, &v5_service_key);
+     if (v4_service_key.contents)
+	 krb5_free_keyblock_contents(context, &v4_service_key);
+     return ret;
+}
+
+/*
+ * afs_return_v4: a predicate to determine whether we want to try
+ * using the afs krb5 encrypted part encoding or whether we  just
+ * return krb4.  Takes a principal, and checks the configuration file.
+ */
+static krb5_error_code 
+afs_return_v4 (krb5_context context, const krb5_principal princ,
+	       int *use_v5)
+{
+    krb5_error_code ret;
+    char *unparsed_name;
+    char *cp;
+    krb5_data realm;
+    assert(use_v5 != NULL);
+    ret = krb5_unparse_name(context, princ, &unparsed_name);
+        if (ret != 0)
+	return ret;
+/* Trim out trailing realm component into separate string.*/
+    for (cp = unparsed_name; *cp != '\0'; cp++) {
+	if (*cp == '\\') {
+	    cp++; /* We trust unparse_name not to leave a singleton
+		   * backslash*/
+	    continue;
+	}
+	if (*cp == '@') {
+	    *cp = '\0';
+	    realm.data = cp+1;
+	    realm.length = strlen((char *) realm.data);
+	    	    break;
+	}
+    }
+     krb5_appdefault_boolean(context, "afs_krb5",
+				  &realm, unparsed_name, 1,
+				  use_v5);
+    krb5_free_unparsed_name(context, unparsed_name);
+    return ret;
+}
