@@ -1,11 +1,11 @@
 /*
  * Copyright 1993 OpenVision Technologies, Inc., All Rights Reserved
  *
- * $Header: /cvs/krbdev/krb5/src/lib/kadm5/srv/svr_principal.c,v 1.19.4.2 2001/10/25 19:24:15 tlyu Exp $
+ * $Header: /cvs/krbdev/krb5/src/lib/kadm5/srv/svr_principal.c,v 1.26.2.1 2003/09/02 18:58:56 lxs Exp $
  */
 
 #if !defined(lint) && !defined(__CODECENTER__)
-static char *rcsid = "$Header: /cvs/krbdev/krb5/src/lib/kadm5/srv/svr_principal.c,v 1.19.4.2 2001/10/25 19:24:15 tlyu Exp $";
+static char *rcsid = "$Header: /cvs/krbdev/krb5/src/lib/kadm5/srv/svr_principal.c,v 1.26.2.1 2003/09/02 18:58:56 lxs Exp $";
 #endif
 
 #include	<sys/types.h>
@@ -19,6 +19,9 @@ static char *rcsid = "$Header: /cvs/krbdev/krb5/src/lib/kadm5/srv/svr_principal.
 #include	"server_internal.h"
 #include	<stdarg.h>
 #include	<stdlib.h>
+#ifdef USE_PASSWORD_SERVER
+#include	<sys/wait.h>
+#endif
 
 extern	krb5_principal	    master_princ;
 extern	krb5_principal	    hist_princ;
@@ -170,8 +173,8 @@ kadm5_create_principal_3(void *server_handle,
 		return ret;
 	}
     }
-    if (ret = passwd_check(handle, password, (mask & KADM5_POLICY),
-			   &polent, entry->principal)) {
+    if ((ret = passwd_check(handle, password, (mask & KADM5_POLICY),
+			    &polent, entry->principal))) {
 	if (mask & KADM5_POLICY)
 	     (void) kadm5_free_policy_ent(handle->lhandle, &polent);
 	return ret;
@@ -181,10 +184,10 @@ kadm5_create_principal_3(void *server_handle,
      * "defaults" for fields that were not specified by the
      * mask.
      */
-    if (ret = krb5_timeofday(handle->context, &now)) {
-	if (mask & KADM5_POLICY)
-	     (void) kadm5_free_policy_ent(handle->lhandle, &polent);
-	return ret;
+    if ((ret = krb5_timeofday(handle->context, &now))) {
+	 if (mask & KADM5_POLICY)
+	      (void) kadm5_free_policy_ent(handle->lhandle, &polent);
+	 return ret;
     }
 
     kdb.magic = KRB5_KDB_MAGIC_NUMBER;
@@ -228,28 +231,28 @@ kadm5_create_principal_3(void *server_handle,
        to free the entire kdb entry, and that will try to free the
        principal. */
 
-    if (ret = krb5_copy_principal(handle->context,
-				  entry->principal, &(kdb.princ))) {
+    if ((ret = krb5_copy_principal(handle->context,
+				   entry->principal, &(kdb.princ)))) {
 	if (mask & KADM5_POLICY)
 	     (void) kadm5_free_policy_ent(handle->lhandle, &polent);
 	return(ret);
     }
 
-    if (ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now)) {
-	krb5_dbe_free_contents(handle->context, &kdb);
-	if (mask & KADM5_POLICY)
+    if ((ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now))) {
+         krb5_dbe_free_contents(handle->context, &kdb);
+	 if (mask & KADM5_POLICY)
 	     (void) kadm5_free_policy_ent(handle->lhandle, &polent);
-	return(ret);
+	 return(ret);
     }
 
     /* initialize the keys */
 
-    if (ret = krb5_dbe_cpw(handle->context, &master_keyblock,
-			   n_ks_tuple?ks_tuple:handle->params.keysalts,
-			   n_ks_tuple?n_ks_tuple:handle->params.num_keysalts,
-			   password,
-			   (mask & KADM5_KVNO)?entry->kvno:1,
-			   FALSE, &kdb)) {
+    if ((ret = krb5_dbe_cpw(handle->context, &master_keyblock,
+			    n_ks_tuple?ks_tuple:handle->params.keysalts,
+			    n_ks_tuple?n_ks_tuple:handle->params.num_keysalts,
+			    password,
+			    (mask & KADM5_KVNO)?entry->kvno:1,
+			    FALSE, &kdb))) {
 	krb5_dbe_free_contents(handle->context, &kdb);
 	if (mask & KADM5_POLICY)
 	     (void) kadm5_free_policy_ent(handle->lhandle, &polent);
@@ -346,7 +349,7 @@ kadm5_delete_principal(void *server_handle, krb5_principal principal)
     if (principal == NULL)
 	return EINVAL;
 
-    if (ret = kdb_get_entry(handle, principal, &kdb, &adb))
+    if ((ret = kdb_get_entry(handle, principal, &kdb, &adb)))
 	return(ret);
 
     if ((adb.aux_attributes & KADM5_POLICY)) {
@@ -362,9 +365,9 @@ kadm5_delete_principal(void *server_handle, krb5_principal principal)
 		return(ret);
 	    }
 	}
-	if (ret = kadm5_free_policy_ent(handle->lhandle, &polent)) {
-	    kdb_free_entry(handle, &kdb, &adb);
-	    return ret;
+	if ((ret = kadm5_free_policy_ent(handle->lhandle, &polent))) {
+	     kdb_free_entry(handle, &kdb, &adb);
+	     return ret;
 	}
     }
 
@@ -383,7 +386,7 @@ kadm5_modify_principal(void *server_handle,
     kadm5_policy_ent_rec    npol, opol;
     int			    have_npol = 0, have_opol = 0;
     krb5_db_entry	    kdb;
-    krb5_tl_data	    *tl_data_orig, *tl_data_tail;
+    krb5_tl_data	    *tl_data_orig;
     osa_princ_ent_rec	    adb;
     kadm5_server_handle_t handle = server_handle;
 
@@ -410,7 +413,8 @@ kadm5_modify_principal(void *server_handle,
 	 }
     }
 
-    if (ret = kdb_get_entry(handle, entry->principal, &kdb, &adb))
+    ret = kdb_get_entry(handle, entry->principal, &kdb, &adb);
+    if (ret)
 	return(ret);
 
     /*
@@ -465,12 +469,13 @@ kadm5_modify_principal(void *server_handle,
 
 	 /* set pw_max_life based on new policy */
 	 if (npol.pw_max_life) {
-	      if (ret = krb5_dbe_lookup_last_pwd_change(handle->context, &kdb,
-							&(kdb.pw_expiration)))
-		   goto done;
-	      kdb.pw_expiration += npol.pw_max_life;
+	     ret = krb5_dbe_lookup_last_pwd_change(handle->context, &kdb,
+						   &(kdb.pw_expiration));
+	     if (ret)
+		 goto done;
+	     kdb.pw_expiration += npol.pw_max_life;
 	 } else {
-	      kdb.pw_expiration = 0;
+	     kdb.pw_expiration = 0;
 	 }
     }
 
@@ -610,7 +615,8 @@ kadm5_rename_principal(void *server_handle,
     }
 
     krb5_free_principal(handle->context, kdb.princ);
-    if (ret = krb5_copy_principal(handle->context, target, &kdb.princ)) {
+    ret = krb5_copy_principal(handle->context, target, &kdb.princ);
+    if (ret) {
 	kdb.princ = NULL; /* so freeing the dbe doesn't lose */
 	goto done;
     }
@@ -696,17 +702,19 @@ kadm5_get_principal(void *server_handle, krb5_principal principal,
     /* this is a little non-sensical because the function returns two */
     /* values that must be checked separately against the mask */
     if ((mask & KADM5_MOD_NAME) || (mask & KADM5_MOD_TIME)) {
-	 if (ret = krb5_dbe_lookup_mod_princ_data(handle->context, &kdb,
-						  &(entry->mod_date),
-						  &(entry->mod_name))) {
-	      goto done;
-	 }
-	 if (! (mask & KADM5_MOD_TIME))
-	      entry->mod_date = 0;
-	 if (! (mask & KADM5_MOD_NAME)) {
-	      krb5_free_principal(handle->context, entry->principal);
-	      entry->principal = NULL;
-	 }
+	ret = krb5_dbe_lookup_mod_princ_data(handle->context, &kdb,
+					     &(entry->mod_date), 
+					     &(entry->mod_name));
+	if (ret) {
+	    goto done;
+	}
+	
+	if (! (mask & KADM5_MOD_TIME))
+	    entry->mod_date = 0;
+	if (! (mask & KADM5_MOD_NAME)) {
+	    krb5_free_principal(handle->context, entry->principal);
+	    entry->principal = NULL;
+	}
     }
 
     if (mask & KADM5_ATTRIBUTES)
@@ -737,7 +745,7 @@ kadm5_get_principal(void *server_handle, krb5_principal principal,
 	 if (mask & KADM5_FAIL_AUTH_COUNT)
 	      entry->fail_auth_count = kdb.fail_auth_count;
 	 if (mask & KADM5_TL_DATA) {
-	      krb5_tl_data td, *tl, *tl2;
+	      krb5_tl_data *tl, *tl2;
 
 	      entry->tl_data = NULL;
 	      
@@ -769,9 +777,10 @@ kadm5_get_principal(void *server_handle, krb5_principal principal,
 		      entry->key_data = NULL;
 
 	      for (i = 0; i < entry->n_key_data; i++)
-		   if (ret = krb5_copy_key_data_contents(handle->context,
-							 &kdb.key_data[i],
-							 &entry->key_data[i]))
+		  ret = krb5_copy_key_data_contents(handle->context,
+						    &kdb.key_data[i],
+						    &entry->key_data[i]);
+		   if (ret)
 			goto done;
 	 }
     }
@@ -848,37 +857,38 @@ static kadm5_ret_t
 check_pw_reuse(krb5_context context,
 	       krb5_keyblock *hist_keyblock,
 	       int n_new_key_data, krb5_key_data *new_key_data,
-	       int n_pw_hist_data, osa_pw_hist_ent *pw_hist_data)
+	       unsigned int n_pw_hist_data, osa_pw_hist_ent *pw_hist_data)
 {
     int x, y, z;
     krb5_keyblock newkey, histkey;
     krb5_error_code ret;
 
     for (x = 0; x < n_new_key_data; x++) {
-	 if (ret = krb5_dbekd_decrypt_key_data(context,
-					       &master_keyblock,
-					       &(new_key_data[x]),
-					       &newkey, NULL))
+	ret = krb5_dbekd_decrypt_key_data(context,
+					  &master_keyblock,
+					  &(new_key_data[x]),
+					  &newkey, NULL);
+	if (ret)
 	    return(ret);
 	for (y = 0; y < n_pw_hist_data; y++) {
 	     for (z = 0; z < pw_hist_data[y].n_key_data; z++) {
-		  if (ret =
-		      krb5_dbekd_decrypt_key_data(context,
-						  hist_keyblock,
-						  &pw_hist_data[y].key_data[z],
-						  &histkey, NULL))
-		       return(ret);		
-		  
-		  if ((newkey.length == histkey.length) &&
-		      (newkey.enctype == histkey.enctype) &&
-		      (memcmp(newkey.contents, histkey.contents,
-			      histkey.length) == 0)) {
-		       krb5_free_keyblock_contents(context, &histkey);
-		       krb5_free_keyblock_contents(context, &newkey);
-		       
-		       return(KADM5_PASS_REUSE);
-		  }
-		  krb5_free_keyblock_contents(context, &histkey);
+		 ret = krb5_dbekd_decrypt_key_data(context,
+						   hist_keyblock,
+						   &pw_hist_data[y].key_data[z],
+						   &histkey, NULL);
+		 if (ret)
+		     return(ret);		
+		 
+		 if ((newkey.length == histkey.length) &&
+		     (newkey.enctype == histkey.enctype) &&
+		     (memcmp(newkey.contents, histkey.contents,
+			     histkey.length) == 0)) {
+		     krb5_free_keyblock_contents(context, &histkey);
+		     krb5_free_keyblock_contents(context, &newkey);
+		     
+		     return(KADM5_PASS_REUSE);
+		 }
+		 krb5_free_keyblock_contents(context, &histkey);
 	     }
 	}
 	krb5_free_keyblock_contents(context, &newkey);
@@ -907,6 +917,7 @@ check_pw_reuse(krb5_context context,
  * in hist_key, and added to hist->key_data.  hist->n_key_data is
  * set to n_key_data.
  */
+static
 int create_history_entry(krb5_context context, int n_key_data,
 			 krb5_key_data *key_data, osa_pw_hist_ent *hist)
 {
@@ -920,26 +931,30 @@ int create_history_entry(krb5_context context, int n_key_data,
      memset(hist->key_data, 0, n_key_data*sizeof(krb5_key_data));
 
      for (i = 0; i < n_key_data; i++) {
-	  if (ret = krb5_dbekd_decrypt_key_data(context,
-						&master_keyblock,
-						&key_data[i],
-						&key, &salt))
-	       return ret;
-	  if (ret = krb5_dbekd_encrypt_key_data(context,
-						&hist_key,
-						&key, &salt,
-						key_data[i].key_data_kvno,
-						&hist->key_data[i]))
-	       return ret;
-	  krb5_free_keyblock_contents(context, &key);
-	  /* krb5_free_keysalt(context, &salt); */
+	 ret = krb5_dbekd_decrypt_key_data(context,
+					   &master_keyblock,
+					   &key_data[i],
+					   &key, &salt);
+	 if (ret)
+	     return ret;
+
+	 ret = krb5_dbekd_encrypt_key_data(context, &hist_key,
+					   &key, &salt,
+					   key_data[i].key_data_kvno,
+					   &hist->key_data[i]);
+	 if (ret)
+	     return ret;
+	 
+	 krb5_free_keyblock_contents(context, &key);
+	 /* krb5_free_keysalt(context, &salt); */
      }
 
      hist->n_key_data = n_key_data;
      return 0;
 }
 
-int free_history_entry(krb5_context context, osa_pw_hist_ent *hist)
+static
+void free_history_entry(krb5_context context, osa_pw_hist_ent *hist)
 {
      int i;
 
@@ -980,8 +995,8 @@ static kadm5_ret_t add_to_history(krb5_context context,
 				  kadm5_policy_ent_t pol,
 				  osa_pw_hist_ent *pw)
 {
-     osa_pw_hist_ent hist, *histp;
-     int ret, i;
+     osa_pw_hist_ent *histp;
+     int i;
 
      /* A history of 1 means just check the current password */
      if (pol->pw_history_num == 1)
@@ -1053,6 +1068,105 @@ static kadm5_ret_t add_to_history(krb5_context context,
 }
 #undef KADM_MOD
 
+#ifdef USE_PASSWORD_SERVER
+
+/* FIXME: don't use global variable for this */
+krb5_boolean use_password_server = 0;
+
+static krb5_boolean
+kadm5_use_password_server (void)
+{
+    return use_password_server;
+}
+
+void
+kadm5_set_use_password_server (void)
+{
+    use_password_server = 1;
+}
+
+/*
+ * kadm5_launch_task () runs a program (task_path) to synchronize the 
+ * Apple password server with the Kerberos database.  Password server
+ * programs can receive arguments on the command line (task_argv)
+ * and a block of data via stdin (data_buffer).
+ *
+ * Because a failure to communicate with the tool results in the
+ * password server falling out of sync with the database,
+ * kadm5_launch_task() always fails if it can't talk to the tool.
+ */
+
+static kadm5_ret_t
+kadm5_launch_task (krb5_context context,
+                   const char *task_path, char * const task_argv[],
+                   const char *data_buffer) 
+{
+    kadm5_ret_t ret = 0;
+    int data_pipe[2];
+    
+    if (data_buffer != NULL) {
+        ret = pipe (data_pipe);
+        if (ret) { ret = errno; }
+    }
+
+    if (!ret) {
+        pid_t pid = fork ();
+        if (pid == -1) {
+            ret = errno;
+        } else if (pid == 0) {
+            /* The child: */
+            
+            if (data_buffer != NULL) {
+                if (dup2 (data_pipe[0], STDIN_FILENO) == -1) {
+                    _exit (1);
+                }
+            } else {
+                close (data_pipe[0]);
+            }
+
+            close (data_pipe[1]);
+            
+            execv (task_path, task_argv);
+            
+            _exit (1); /* Fail if execv fails */
+        } else {
+            /* The parent: */
+            int status;
+                       
+            if (data_buffer != NULL) {
+                /* Write out the buffer to the child */
+                if (krb5_net_write (context, data_pipe[1],
+                                    data_buffer, strlen (data_buffer)) < 0) {
+                    /* kill the child to make sure waitpid() won't hang later */
+                    ret = errno;
+                    kill (pid, SIGKILL);
+                }
+            }
+
+            close (data_buffer[0]);
+            close (data_buffer[1]);
+
+            waitpid (pid, &status, 0);
+
+            if (!ret) {
+                if (WIFEXITED (status)) {
+                    /* child read password and exited.  Check the return value. */
+                    if ((WEXITSTATUS (status) != 0) && (WEXITSTATUS (status) != 252)) {
+                       ret = KRB5KDC_ERR_POLICY; /* password change rejected */
+                    }
+                } else {
+                    /* child read password but crashed or was killed */
+                    ret = KRB5KRB_ERR_GENERIC; /* FIXME: better error */
+                }
+            }
+        }
+    }
+
+    return ret;
+}
+
+#endif
+
 kadm5_ret_t
 kadm5_chpass_principal(void *server_handle,
 			    krb5_principal principal, char *password)
@@ -1072,7 +1186,7 @@ kadm5_chpass_principal_3(void *server_handle,
     kadm5_policy_ent_rec	pol;
     osa_princ_ent_rec		adb;
     krb5_db_entry		kdb, kdb_save;
-    int				ret, ret2, last_pwd, i, hist_added;
+    int				ret, ret2, last_pwd, hist_added;
     int				have_pol = 0;
     kadm5_server_handle_t	handle = server_handle;
     osa_pw_hist_ent		hist;
@@ -1107,24 +1221,27 @@ kadm5_chpass_principal_3(void *server_handle,
 			    KADM5_POLICY, &pol, principal)))
 	 goto done;
 
-    if (ret = krb5_dbe_cpw(handle->context, &master_keyblock,
-			   n_ks_tuple?ks_tuple:handle->params.keysalts,
-			   n_ks_tuple?n_ks_tuple:handle->params.num_keysalts,
-			   password, 0 /* increment kvno */,
-			   keepold, &kdb))
+    ret = krb5_dbe_cpw(handle->context, &master_keyblock,
+		       n_ks_tuple?ks_tuple:handle->params.keysalts,
+		       n_ks_tuple?n_ks_tuple:handle->params.num_keysalts,
+		       password, 0 /* increment kvno */,
+		       keepold, &kdb);
+    if (ret)
 	goto done;
 
     kdb.attributes &= ~KRB5_KDB_REQUIRES_PWCHANGE;
 
-    if (ret = krb5_timeofday(handle->context, &now))
+    ret = krb5_timeofday(handle->context, &now);
+    if (ret)
 	 goto done;
     
     if ((adb.aux_attributes & KADM5_POLICY)) {
        /* the policy was loaded before */
 
-	if (ret = krb5_dbe_lookup_last_pwd_change(handle->context,
-						  &kdb, &last_pwd))
-	     goto done;
+	ret = krb5_dbe_lookup_last_pwd_change(handle->context,
+					      &kdb, &last_pwd);
+	if (ret)
+	    goto done;
 
 #if 0
 	 /*
@@ -1140,16 +1257,17 @@ kadm5_chpass_principal_3(void *server_handle,
 	}
 #endif
 
-	if (ret = create_history_entry(handle->context,
-				       kdb_save.n_key_data,
-				       kdb_save.key_data, &hist))
-	     goto done;
+	ret = create_history_entry(handle->context,
+				   kdb_save.n_key_data,
+				   kdb_save.key_data, &hist);
+	if (ret)
+	    goto done;
 
-	if (ret = check_pw_reuse(handle->context,
-				 &hist_key,
-				 kdb.n_key_data, kdb.key_data,
-				 1, &hist))
-	     goto done;
+	ret = check_pw_reuse(handle->context, &hist_key,
+			     kdb.n_key_data, kdb.key_data,
+			     1, &hist);
+	if (ret)
+	    goto done;
 	 
 	if (pol.pw_history_num > 1) {
 	    if (adb.admin_history_kvno != hist_kvno) {
@@ -1157,14 +1275,15 @@ kadm5_chpass_principal_3(void *server_handle,
 		goto done;
 	    }
 
-	    if (ret = check_pw_reuse(handle->context,
-				     &hist_key,
-				     kdb.n_key_data, kdb.key_data,
-				     adb.old_key_len, adb.old_keys))
+	    ret = check_pw_reuse(handle->context, &hist_key,
+				 kdb.n_key_data, kdb.key_data,
+				 adb.old_key_len, adb.old_keys);
+	    if (ret)
 		goto done;
 
-	    if (ret = add_to_history(handle->context, &adb, &pol, &hist))
-		 goto done;
+	    ret = add_to_history(handle->context, &adb, &pol, &hist);
+	    if (ret)
+		goto done;
 	    hist_added = 1;
        }
 
@@ -1176,7 +1295,44 @@ kadm5_chpass_principal_3(void *server_handle,
 	kdb.pw_expiration = 0;
     }
 
-    if (ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now))
+#ifdef USE_PASSWORD_SERVER
+    if (kadm5_use_password_server () &&
+        (krb5_princ_size (handle->context, principal) == 1)) {
+        krb5_data *princ = krb5_princ_component (handle->context, principal, 0);
+        const char *path = "/usr/sbin/mkpassdb";
+        char *argv[] = { "mkpassdb", "-setpassword", NULL, NULL };
+        char *pstring = NULL;
+        char pwbuf[256];
+        int pwlen = strlen (password);
+
+        if (pwlen > 254) pwlen = 254;
+        strncpy (pwbuf, password, pwlen);
+        pwbuf[pwlen] = '\n';
+        pwbuf[pwlen + 1] = '\0';
+
+        if (!ret) {
+            pstring = malloc ((princ->length + 1) * sizeof (char));
+            if (pstring == NULL) { ret = errno; }
+        }
+
+        if (!ret) {
+            memcpy (pstring, princ->data, princ->length);
+            pstring [princ->length] = '\0';
+            argv[2] = pstring;
+
+            ret = kadm5_launch_task (handle->context, path, argv, pwbuf);
+        }
+        
+        if (pstring != NULL)
+            free (pstring);
+        
+        if (ret)
+            goto done;
+    }
+#endif
+
+    ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now);
+    if (ret)
 	goto done;
 
     if ((ret = kdb_put_entry(handle, &kdb, &adb)))
@@ -1221,7 +1377,6 @@ kadm5_randkey_principal_3(void *server_handle,
     krb5_int32			now;
     kadm5_policy_ent_rec	pol;
     krb5_key_data		*key_data;
-    krb5_keyblock		*keyblock;
     int				ret, last_pwd, have_pol = 0;
     kadm5_server_handle_t	handle = server_handle;
 
@@ -1240,16 +1395,18 @@ kadm5_randkey_principal_3(void *server_handle,
     if ((ret = kdb_get_entry(handle, principal, &kdb, &adb)))
        return(ret);
 
-    if (ret = krb5_dbe_crk(handle->context, &master_keyblock,
-			   n_ks_tuple?ks_tuple:handle->params.keysalts,
-			   n_ks_tuple?n_ks_tuple:handle->params.num_keysalts,
-			   keepold,
-			   &kdb))
-       goto done;
+    ret = krb5_dbe_crk(handle->context, &master_keyblock,
+		       n_ks_tuple?ks_tuple:handle->params.keysalts,
+		       n_ks_tuple?n_ks_tuple:handle->params.num_keysalts,
+		       keepold,
+		       &kdb);
+    if (ret)
+	goto done;
 
     kdb.attributes &= ~KRB5_KDB_REQUIRES_PWCHANGE;
 
-    if (ret = krb5_timeofday(handle->context, &now))
+    ret = krb5_timeofday(handle->context, &now);
+    if (ret)
 	goto done;
 
     if ((adb.aux_attributes & KADM5_POLICY)) {
@@ -1258,8 +1415,9 @@ kadm5_randkey_principal_3(void *server_handle,
 	   goto done;
 	have_pol = 1;
 
-	if (ret = krb5_dbe_lookup_last_pwd_change(handle->context,
-						  &kdb, &last_pwd))
+	ret = krb5_dbe_lookup_last_pwd_change(handle->context,
+					      &kdb, &last_pwd);
+	if (ret)
 	     goto done;
 
 #if 0
@@ -1282,10 +1440,10 @@ kadm5_randkey_principal_3(void *server_handle,
 		goto done;
 	    }
 
-	    if (ret = check_pw_reuse(handle->context,
-				     &hist_key,
-				     kdb.n_key_data, kdb.key_data,
-				     adb.old_key_len, adb.old_keys))
+	    ret = check_pw_reuse(handle->context, &hist_key,
+				 kdb.n_key_data, kdb.key_data,
+				 adb.old_key_len, adb.old_keys);
+	    if (ret)
 		goto done;
 	}
 	if (pol.pw_max_life)
@@ -1296,26 +1454,29 @@ kadm5_randkey_principal_3(void *server_handle,
 	kdb.pw_expiration = 0;
     }
 
-    if (ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now))
+    ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now);
+    if (ret)
 	 goto done;
 
     if (keyblocks) {
 	 if (handle->api_version == KADM5_API_VERSION_1) {
 	      /* Version 1 clients will expect to see a DES_CRC enctype. */
-	      if (ret = krb5_dbe_find_enctype(handle->context, &kdb,
-					      ENCTYPE_DES_CBC_CRC,
-					      -1, -1, &key_data))
-		   goto done;
-
-	      if (ret = decrypt_key_data(handle->context, 1, key_data,
-					 keyblocks, NULL))
-		   goto done;
+	     ret = krb5_dbe_find_enctype(handle->context, &kdb,
+					 ENCTYPE_DES_CBC_CRC,
+					 -1, -1, &key_data);
+	     if (ret)
+		 goto done;
+	     
+	     ret = decrypt_key_data(handle->context, 1, key_data,
+				     keyblocks, NULL);
+	     if (ret)
+		 goto done;
 	 } else {
-	      ret = decrypt_key_data(handle->context,
+	     ret = decrypt_key_data(handle->context,
 				     kdb.n_key_data, kdb.key_data,
 				     keyblocks, n_keys);
-	      if (ret)
-		   goto done;
+	     if (ret)
+		 goto done;
 	 }
     }	 
     
@@ -1347,10 +1508,11 @@ kadm5_setv4key_principal(void *server_handle,
     osa_princ_ent_rec		adb;
     krb5_int32			now;
     kadm5_policy_ent_rec	pol;
-    krb5_key_data		*key_data;
     krb5_keysalt		keysalt;
-    int				i, kvno, ret, last_pwd, have_pol = 0;
-    int				deskeys;
+    int				i, kvno, ret, have_pol = 0;
+#if 0
+    int                         last_pwd;
+#endif
     kadm5_server_handle_t	handle = server_handle;
 
     CHECK_HANDLE(server_handle);
@@ -1385,17 +1547,17 @@ kadm5_setv4key_principal(void *server_handle,
     keysalt.data.length = 0;
     keysalt.data.data = NULL;
 
-    if (ret = krb5_dbekd_encrypt_key_data(handle->context,
-					  &master_keyblock,
-					  keyblock, &keysalt,
-					  kvno + 1,
-					  kdb.key_data)) {
+    ret = krb5_dbekd_encrypt_key_data(handle->context, &master_keyblock,
+				      keyblock, &keysalt, kvno + 1,
+				      kdb.key_data);
+    if (ret) {
 	goto done;
     }
 
     kdb.attributes &= ~KRB5_KDB_REQUIRES_PWCHANGE;
 
-    if (ret = krb5_timeofday(handle->context, &now))
+    ret = krb5_timeofday(handle->context, &now);
+    if (ret)
 	goto done;
 
     if ((adb.aux_attributes & KADM5_POLICY)) {
@@ -1446,7 +1608,8 @@ kadm5_setv4key_principal(void *server_handle,
 	kdb.pw_expiration = 0;
     }
 
-    if (ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now))
+    ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now);
+    if (ret)
 	 goto done;
 
     if ((ret = kdb_put_entry(handle, &kdb, &adb)))
@@ -1487,7 +1650,10 @@ kadm5_setkey_principal_3(void *server_handle,
     kadm5_policy_ent_rec	pol;
     krb5_key_data		*old_key_data;
     int				n_old_keys;
-    int				i, j, kvno, ret, last_pwd, have_pol = 0;
+    int				i, j, kvno, ret, have_pol = 0;
+#if 0
+    int                         last_pwd;
+#endif
     kadm5_server_handle_t	handle = server_handle;
     krb5_boolean		similar;
     krb5_keysalt		keysalt;
@@ -1503,21 +1669,22 @@ kadm5_setkey_principal_3(void *server_handle,
 
     for (i = 0; i < n_keys; i++) {
 	for (j = i+1; j < n_keys; j++) {
-	    if (ret = krb5_c_enctype_compare(handle->context,
-					     keyblocks[i].enctype,
-					     keyblocks[j].enctype,
-					     &similar))
+	    if ((ret = krb5_c_enctype_compare(handle->context,
+					      keyblocks[i].enctype,
+					      keyblocks[j].enctype,
+					      &similar)))
 		return(ret);
-	    if (similar)
+	    if (similar) {
 		if (n_ks_tuple) {
 		    if (ks_tuple[i].ks_salttype == ks_tuple[j].ks_salttype)
 			return KADM5_SETKEY_DUP_ENCTYPES;
 		} else
 		    return KADM5_SETKEY_DUP_ENCTYPES;
+	    }
 	}
     }
 
-    if (n_ks_tuple != n_keys)
+    if (n_ks_tuple && n_ks_tuple != n_keys)
 	return KADM5_SETKEY3_ETYPE_MISMATCH;
 
     if ((ret = kdb_get_entry(handle, principal, &kdb, &adb)))
@@ -1577,7 +1744,7 @@ kadm5_setkey_principal_3(void *server_handle,
     /* assert(kdb.n_key_data == n_keys + n_old_keys) */
     kdb.attributes &= ~KRB5_KDB_REQUIRES_PWCHANGE;
 
-    if (ret = krb5_timeofday(handle->context, &now))
+    if ((ret = krb5_timeofday(handle->context, &now)))
 	goto done;
 
     if ((adb.aux_attributes & KADM5_POLICY)) {
@@ -1628,8 +1795,8 @@ kadm5_setkey_principal_3(void *server_handle,
 	kdb.pw_expiration = 0;
     }
 
-    if (ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now))
-	 goto done;
+    if ((ret = krb5_dbe_update_last_pwd_change(handle->context, &kdb, now)))
+        goto done;
 
     if ((ret = kdb_put_entry(handle, &kdb, &adb)))
 	goto done;
@@ -1662,10 +1829,10 @@ static int decrypt_key_data(krb5_context context,
      memset((char *) keys, 0, n_key_data*sizeof(krb5_keyblock));
 
      for (i = 0; i < n_key_data; i++) {
-	  if (ret = krb5_dbekd_decrypt_key_data(context,
-						&master_keyblock,
-						&key_data[i], 
-						&keys[i], NULL)) {
+          ret = krb5_dbekd_decrypt_key_data(context, &master_keyblock,
+					    &key_data[i], 
+					    &keys[i], NULL);
+	  if (ret) {
 
 	       memset((char *) keys, 0, n_key_data*sizeof(krb5_keyblock));
 	       free(keys);
@@ -1729,13 +1896,13 @@ kadm5_ret_t kadm5_decrypt_key(void *server_handle,
     /* find_enctype only uses these two fields */
     dbent.n_key_data = entry->n_key_data;
     dbent.key_data = entry->key_data;
-    if (ret = krb5_dbe_find_enctype(handle->context, &dbent, ktype,
-				    stype, kvno, &key_data))
+    if ((ret = krb5_dbe_find_enctype(handle->context, &dbent, ktype,
+				    stype, kvno, &key_data)))
 	 return ret;
 
-    if (ret = krb5_dbekd_decrypt_key_data(handle->context,
-					  &master_keyblock, key_data,
-					  keyblock, keysalt))
+    if ((ret = krb5_dbekd_decrypt_key_data(handle->context,
+					   &master_keyblock, key_data,
+					   keyblock, keysalt)))
 	 return ret;
 
     /*
