@@ -37,7 +37,7 @@
 
 apiCB *gCntrlBlock = NULL;
 
-#if defined(_MSDOS) || defined(_WIN32)
+#if defined(_WIN32)
 #include "winccld.h"	
 #endif
 
@@ -46,7 +46,7 @@ apiCB *gCntrlBlock = NULL;
 #endif
 
 #ifdef DEBUG
-#if defined(_MSDOS) || defined(_WIN32)
+#if defined(_WIN32)
 #include <io.h>
 #define SHOW_DEBUG(buf)   MessageBox((HWND)NULL, (buf), "ccapi debug", MB_OK)
 #endif
@@ -80,7 +80,7 @@ krb5_cc_ops krb5_cc_stdcc_ops = {
       krb5_stdcc_set_flags,
 };
 
-#if defined(_MSDOS) || defined(_WIN32)
+#if defined(_WIN32)
 /*
  * cache_changed be called after the cache changes.
  * A notification message is is posted out to all top level
@@ -97,13 +97,13 @@ void cache_changed()
 
 	PostMessage(HWND_BROADCAST, message, 0, 0);
 }
-#else /* _MSDOS || _WIN32 */
+#else /* _WIN32 */
 
 void cache_changed()
 {
 	return;
 }
-#endif /* _MSDOS || _WIN32 */
+#endif /* _WIN32 */
 
 struct err_xlate
 {
@@ -265,7 +265,7 @@ krb5_error_code KRB5_CALLCONV  krb5_stdcc_resolve
 	stdccCacheDataPtr	ccapi_data = NULL;
 	int 			err;
 	krb5_error_code		retval;
-	char 			*cName;
+	char 			*cName = NULL;
 	
 	if ((retval = stdcc_setup(context, NULL)))
 		return retval;
@@ -549,7 +549,9 @@ krb5_error_code KRB5_CALLCONV krb5_stdcc_end_seq_get
 	krb5_error_code		retval;
 	stdccCacheDataPtr	ccapi_data = NULL;
 	int			err;
+#ifndef CC_API_VER2
 	cred_union 		*credU = NULL;
+#endif
 
 	ccapi_data = id->data;
 	
@@ -654,10 +656,9 @@ krb5_stdcc_destroy (krb5_context context, krb5_ccache id)
  *
  * - return the name of the named cache
  */
-char * KRB5_CALLCONV krb5_stdcc_get_name 
+const char * KRB5_CALLCONV krb5_stdcc_get_name 
         (krb5_context context, krb5_ccache id )
 {
-	char *name = NULL;
 	stdccCacheDataPtr	ccapi_data = id->data;
 
 	if (!ccapi_data)
