@@ -1,5 +1,5 @@
 /*
- * Copyright 1997 by Massachusetts Institute of Technology
+ * Copyright 1997,2000,2001 by Massachusetts Institute of Technology
  * 
  * Copyright 1987, 1988 by MIT Student Information Processing Board
  *
@@ -27,55 +27,61 @@
 #include "com_err.h"
 #include "error_table.h"
 
-#if defined(macintosh) || (defined(__MACH__) && defined(__APPLE__))
-    #include <KerberosSupport/KerberosSupport.h>
-    #include <KerberosSupport/ErrorLib.h>
-#endif
-
-#if defined(_MSDOS) || defined(_WIN32)
+#if defined(_WIN32)
 #define HAVE_STRERROR
 #endif
 
-#ifdef macintosh
-#define sys_nerr 100   /* XXX - What is this? */
-#endif
-
 #if !defined(HAVE_STRERROR) && !defined(SYS_ERRLIST_DECLARED)
 extern char const * const sys_errlist[];
-#ifndef macintosh
 extern const int sys_nerr;
 #endif
-#endif
 
 static char buffer[ET_EBUFSIZ];
 
-#if (defined(_MSDOS) || defined(_WIN32) || defined(macintosh) || (defined(__MACH__) && defined(__APPLE__)))
-static struct et_list * _et_list = (struct et_list *) NULL;
+#if (defined(_WIN32) || TARGET_OS_MAC)
+/*@null@*/ static struct et_list * _et_list = (struct et_list *) NULL;
 #else
 /* Old interface compatibility */
-struct et_list * _et_list = (struct et_list *) NULL;
+/*@null@*/ struct et_list * _et_list = (struct et_list *) NULL;
 #endif
 
-KRB5_DLLIMP const char FAR * KRB5_CALLCONV error_message(code)
-	long code;
+/*@null@*//*@only@*/static struct dynamic_et_list * et_list_dynamic;
+
+#ifndef DEBUG_TABLE_LIST
+#define dprintf(X)
+#else
+#define dprintf(X) printf X
+#endif
+
+const char * KRB5_CALLCONV
+error_message(long code)
+    /*@modifies internalState@*/
 {
 	unsigned long offset;
 	unsigned long l_offset;
 	struct et_list *et;
+	struct dynamic_et_list *det;
 	unsigned long table_num;
 	int started = 0;
 	unsigned int divisor = 100;
 	char *cp;
+	const struct error_table *table;
 
 	l_offset = (unsigned long)code & ((1<<ERRCODE_RANGE)-1);
 	offset = l_offset;
 	table_num = ((unsigned long)code - l_offset) & ERRCODE_MAX;
-	if (!table_num) {
+	if (table_num == 0) {
+#ifdef __sgi
+	system_error_code:
+#endif
 		if (code == 0)
 			goto oops;
 
+		/* This could trip if int is 16 bits.  */
+		if ((unsigned long)(int)code != code)
+		    abort ();
 #ifdef HAVE_STRERROR
-		cp = strerror(offset);
+		cp = strerror((int) code);
 		if (cp)
 			return cp;
 		goto oops;
@@ -90,19 +96,51 @@ KRB5_DLLIMP const char FAR * KRB5_CALLCONV error_message(code)
 #endif /* HAVE_SYS_ERRLIST */
 #endif /* HAVE_STRERROR */
 	}
+#ifdef __sgi
+	/* Irix 6.5 uses a much bigger table than other UNIX systems
+	   I've looked at, but the table is sparse.  The sparse
+	   entries start around 500, but sys_nerr is only 152.  */
+	if (code > 0 && code <= 1600)
+	    goto system_error_code;
+#endif
 
-	et = _et_list;
-	while (et) {
+	dprintf (("scanning static list for %x\n", table_num));
+	for (et = _et_list; et != NULL; et = et->next) {
+	    if (et->table == NULL)
+		continue;
+	    dprintf (("\t%x = %s\n", et->table->base & ERRCODE_MAX,
+		      et->table->msgs[0]));
 	    if ((et->table->base & ERRCODE_MAX) == table_num) {
-			/* This is the right table */
-			if (et->table->n_msgs <= offset)
-				break;
-			return(et->table->msgs[offset]);
-		}
-		et = et->next;
+		table = et->table;
+		goto found;
+	    }
 	}
+	dprintf (("scanning dynamic list for %x\n", table_num));
+	for (det = et_list_dynamic; det != NULL; det = det->next) {
+	    dprintf (("\t%x = %s\n", det->table->base & ERRCODE_MAX,
+		      det->table->msgs[0]));
+	    if ((det->table->base & ERRCODE_MAX) == table_num) {
+		table = det->table;
+		goto found;
+	    }
+	}
+	goto no_table_found;
 
-#if defined(_MSDOS) || defined(_WIN32)
+ found:
+	dprintf (("found it!\n"));
+	/* This is the right table */
+
+	/* This could trip if int is 16 bits.  */
+	if ((unsigned long)(unsigned int)offset != offset)
+	    goto no_table_found;
+
+	if (table->n_msgs <= (unsigned int) offset)
+	    goto no_table_found;
+
+	return table->msgs[offset];
+
+ no_table_found:
+#if defined(_WIN32)
 	/*
 	 * WinSock errors exist in the 10000 and 11000 ranges
 	 * but might not appear if WinSock is not initialized
@@ -153,31 +191,23 @@ oops:
 
 #if TARGET_OS_MAC
 	{
-		/* This may be a Mac OS Toolbox error or an MIT Support Library Error.  Ask ErrorLib */
-		if (GetErrorLongString(code, buffer, ET_EBUFSIZ - 1) == noErr) {
-			return buffer;
-		}
-
-#if TARGET_API_MAC_OSX
-		/* ComErr and ErrorLib don't know about this error, ask the system */
+		/* ComErr doesn't know about this error, ask the system */
 		/* Of course there's no way to tell if it knew what error it got */
 		return (strerror (code));
-#endif
-
 	}
 #endif
 	
 	cp = buffer;
 	strcpy(cp, "Unknown code ");
 	cp += sizeof("Unknown code ") - 1;
-	if (table_num) {
-		error_table_name_r(table_num, cp);
-		while (*cp)
+	if (table_num != 0L) {
+		(void) error_table_name_r(table_num, cp);
+		while (*cp != '\0')
 			cp++;
 		*cp++ = ' ';
 	}
 	while (divisor > 1) {
-	    if (started || offset >= divisor) {
+	    if (started != 0 || offset >= divisor) {
 		*cp++ = '0' + offset / divisor;
 		offset %= divisor;
 		started++;
@@ -189,66 +219,49 @@ oops:
 	return(buffer);
 }
 
-
-#ifdef _MSDOS
-/*
- * Win16 applications cannot call malloc while the DLL is being
- * initialized...  To get around this, we pre-allocate an array
- * sufficient to hold several error tables.
- */
-#define PREALLOCATE_ETL 32
-static struct et_list etl[PREALLOCATE_ETL];
-static int etl_used = 0;
-#endif
-
-KRB5_DLLIMP errcode_t KRB5_CALLCONV
-add_error_table(et)
-    const struct error_table FAR * et;
+/*@-incondefs@*/ /* _et_list is global on unix but not in header annotations */
+errcode_t KRB5_CALLCONV
+add_error_table(/*@dependent@*/ const struct error_table * et)
+     /*@modifies _et_list,et_list_dynamic@*/
+/*@=incondefs@*/
 {
-    struct et_list *el = _et_list;
+    struct dynamic_et_list *del;
 
-    while (el) {
-	if (el->table->base == et->base)
-	    return EEXIST;
-	el = el->next;
-    }
-
-#ifdef _MSDOS
-    if (etl_used < PREALLOCATE_ETL)
-	el = &etl[etl_used++];
-    else
-#endif
-	if (!(el = (struct et_list *)malloc(sizeof(struct et_list))))
-	    return ENOMEM;
-
-    el->table = et;
-    el->next = _et_list;
-    _et_list = el;
+    del = (struct dynamic_et_list *)malloc(sizeof(struct dynamic_et_list));
+    if (del == NULL)
+	return errno;
 
+    del->table = et;
+    del->next = et_list_dynamic;
+    et_list_dynamic = del;
     return 0;
 }
 
-KRB5_DLLIMP errcode_t KRB5_CALLCONV
-remove_error_table(et)
-    const struct error_table FAR * et;
+/*@-incondefs@*/ /* _et_list is global on unix but not in header annotations */
+errcode_t KRB5_CALLCONV
+remove_error_table(const struct error_table * et)
+     /*@modifies _et_list,et_list_dynamic@*/
+/*@=incondefs@*/
 {
-    struct et_list *el = _et_list;
-    struct et_list *el2 = 0;
+    struct dynamic_et_list **del;
+    struct et_list **el;
 
-    while (el) {
-	if (el->table->base == et->base) {
-	    if (el2)	/* Not the beginning of the list */
-		el2->next = el->next;
-	    else
-		_et_list = el->next;
-#ifdef _MSDOS
-	    if ((el < etl) || (el > &etl[PREALLOCATE_ETL-1]))
-#endif
-		(void) free(el);
+    /* Remove the first occurrance we can find.  Prefer dynamic
+       entries, but if there are none, check for a static one too.  */
+    for (del = &et_list_dynamic; *del; del = &(*del)->next)
+	if ((*del)->table->base == et->base) {
+	    /*@only@*/ struct dynamic_et_list *old = *del;
+	    *del = old->next;
+	    free (old);
+	    return 0;
+	}
+    for (el = &_et_list; *el; el = &(*el)->next)
+	if ((*el)->table != NULL && (*el)->table->base == et->base) {
+	    struct et_list *old = *el;
+	    *el = old->next;
+	    old->next = NULL;
+	    old->table = NULL;
 	    return 0;
 	}
-	el2 = el;
-	el = el->next;
-    }
     return ENOENT;
 }
